#!/usr/bin/env python

# -----------------------------------------------------------------------------
# This file is part of icepapcms https://github.com/ALBA-Synchrotron/icepapcms
#
# Copyright 2008-2018 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# -----------------------------------------------------------------------------

from .singleton import Singleton
from PyQt5 import QtCore, QtGui, QtWidgets
from icepap import Mode
import logging

from ..gui.messagedialogs import MessageDialogs
from .stormmanager import StormManager
from .configmanager import ConfigManager
from .icepapsystem import IcepapSystem, Location
from .conflict import Conflict
from .icepapsmanager import IcepapsManager
from ..helpers import loggingInfo

__all__ = ['MainManager']


class MainManager(Singleton):
    log = logging.getLogger('{}.MainManager'.format(__name__))

    def __init__(self, form=None):
        pass

    @loggingInfo
    def init(self, *args):
        self.IcepapSystemList = {}

        self._ctrl_icepap = IcepapsManager()
        self._db = StormManager()

        self.dbStatusOK = self._db.dbOK

        self._form = args[0]
        self.locationList = self._db.getAllLocations()
        self.location = None
        self.IcepapSystemList = {}


    @loggingInfo
    def addLocation(self, location_name):
        """ Adds a location in the database """
        try:
            if str(location_name) in self.locationList:
                return False
            location = Location(str(location_name))
            self._db.store(location)
            self._db.commitTransaction()
            self.locationList[str(location_name)] = location
            return True
        except Exception as e:
            self.log.error("Can not add location %s: %s", location_name, e)
            return False


    @loggingInfo
    def deleteLocation(self, location_name):
        """ Deletes a location in the database """
        location = self.locationList[str(location_name)]
        self._db.deleteLocation(location)
        del self.locationList[str(location_name)]
        self.IcepapSystemList = {}


    @loggingInfo
    def changeLocation(self, location):
        """ Close the connection from all the icepaps in one location.
        And gets all the icepaps from the selected location """

        self.location = self.locationList[str(location)]
        self._ctrl_icepap.closeAllConnections()
        self.IcepapSystemList = self._db.getLocationIcepapSystem(str(location))


    @loggingInfo
    def reset(self, form):
        self._ctrl_icepap.reset()
        self._db.reset()
        self.dbStatusOK = self._db.dbOK
        self.location = None
        self.IcepapSystemList = {}
        self._form = form
        self.locationList = self._db.getAllLocations()


    @loggingInfo
    def addIcepapSystem(self, host, port, description=None):
        """ Adds a new Icepap in the current location, the parameters are
        the hostname, port and description, this function checks if the
        icepap is available, the gets all the configuration of all the
        driver and stores all these information in the database """
        QtWidgets.QApplication.instance().setOverrideCursor(
            QtGui.QCursor(QtCore.Qt.WaitCursor))
        try:
            icepap_name = host
            """ *TO-DO STORM review"""
            if icepap_name in self.IcepapSystemList:
                return None
            location = self.location.name
            try:
                port = int(port)
                icepap_system = IcepapSystem(icepap_name, host,
                                             port, location, description)

                # JUST IN CASE A USER WANTS TO RE-ADD A SYSTEM !?!?!?!?
                # A QUANTUM PROBABILITY TO BE ON TWO LOCATIONS AT THE
                # SAME TIME... :-D
                db_icepap_system = self._db.getIcepapSystem(icepap_name)
                if db_icepap_system is not None:
                    MessageDialogs.showErrorMessage(
                        self._form, "Adding Icepap System error",
                        "The icepap system is already in "
                        "location: '%s'" % (db_icepap_system.location_name))
                    QtWidgets.QApplication.instance().restoreOverrideCursor()
                    return None

                self._ctrl_icepap.openConnection(icepap_name, host, port)
                driver_list = self._ctrl_icepap.scanIcepapSystem(icepap_name)
                self._db.addIcepapSystem(icepap_system)
                self.IcepapSystemList[icepap_name] = icepap_system
                self.location.addSystem(icepap_system)
                for driver in list(driver_list.values()):
                    icepap_system.addDriver(driver)
                    self._db.store(driver)
                self._db.commitTransaction()
                QtWidgets.QApplication.instance().restoreOverrideCursor()
                if len(driver_list) == 0:
                    msg = 'The icepap system "{}" has ZERO drivers.\n' \
                          'Please make sure sure that the CAN BUS TERMINATOR' \
                          ' is connected '.format(icepap_name)
                    self.log.warning(msg)
                    MessageDialogs.showWarningMessage(
                        None, "Scanning Icepap Warning", msg)
                return icepap_system
            except Exception as ie:
                msg = 'Could not scan the "{0}" Icepap System.\n' \
                      '1) Please make sure that the system "{0}" is in your ' \
                      'subnetwork.\n' \
                      '2) Run in debug mode to check communication: ' \
                      'icepapcms -d\n' \
                      'Exception: {1}'.format(icepap_name, str(ie))
                self.log.warning(msg)
                MessageDialogs.showErrorMessage(None, "Scanning Icepap Error",
                                                msg)

            except Exception as e:
                self.log.error("Error on adding Icepap System %s:%d: %s",
                               host, port, e)

        except Exception as e:
            self.log.error("Error on adding Icepap System %s:%d: %s",
                           host, port, e)
        QtWidgets.QApplication.instance().restoreOverrideCursor()
        return None

    @loggingInfo
    def deleteIcepapSystem(self, icepap_name):
        """ deletes and Icepap in the database """
        del self.IcepapSystemList[icepap_name]
        self.location.deleteSystem(icepap_name)

    @loggingInfo
    def closeAllConnections(self):
        """ Close all the openned connections to the icepaps """
        self._ctrl_icepap.closeAllConnections()
        return self._db.closeDB()

    @loggingInfo
    def getIcepapSystem(self, icepap_name):
        if icepap_name in self.IcepapSystemList:
            return self.IcepapSystemList[icepap_name]
        return None

    @loggingInfo
    def checkIcepapSystems(self):
        """ Checks if the icepaps for the current location are available over
        the network or not. These function is used to perform automatic
        reconnection """

        changed_list = []
        for icepap_system in list(self.IcepapSystemList.values()):
            connected = self._ctrl_icepap.checkIcepapStatus(icepap_system.name)

            if connected:
                if icepap_system.conflict == Conflict.NO_CONNECTION:
                    icepap_system.conflict = Conflict.NO_CONFLICT
                    changed_list.append(icepap_system)
            else:
                if icepap_system.conflict != Conflict.NO_CONFLICT and \
                        icepap_system.conflict != Conflict.NO_CONNECTION:
                    icepap_system.conflict = Conflict.NO_CONNECTION
                    changed_list.append(icepap_system)
        return changed_list

    @loggingInfo
    def stopIcepap(self, icepap_system):
        """ Close the connection to an icepap. And commits all the changes in
        the database """
        try:
            self._ctrl_icepap.closeConnection(icepap_system.name)
            self._db.commitTransaction()
        except Exception as e:
            msg = "Unexpected error on stop icepap_system {}: " \
                  "{}".format(icepap_system, e)
            self.log.error(msg)
            MessageDialogs.showErrorMessage(self._form, 'Unexpected error',
                                            msg)

    @loggingInfo
    def scanIcepap(self, icepap_system):
        """ Searches for configuration conflicts.
        Returns the conflicts list. That is composed by elements of
        [Conflict code, icepap_system, icepap_driver_addr] """
        QtWidgets.QApplication.instance().setOverrideCursor(
            QtGui.QCursor(QtCore.Qt.WaitCursor))
        icepap_name = icepap_system.name
        conflictsList = []
        try:
            if self._ctrl_icepap.openConnection(icepap_name,
                                                icepap_system.host,
                                                icepap_system.port):
                driver_list = self._ctrl_icepap.scanIcepapSystem(icepap_name,
                                                                 True)
                conflictsList = icepap_system.compareDriverList(driver_list)
            else:
                conflictsList.append([Conflict.NO_CONNECTION, icepap_system,
                                      0])
        except Exception as error:
            msg = 'Could not scan the "{0}" Icepap System.\n' \
                  '1) Please make sure that the system "{0}" is in your ' \
                  'subnetwork.\n' \
                  '2) Run in debug mode to check communication: ' \
                  'icepapcms -d\n' \
                  'Exception: {1}'.format(icepap_name, str(error))
            self.log.warning(msg)
            MessageDialogs.showErrorMessage(None, "Scanning Icepap Error",
                                            msg)

            conflictsList.append([Conflict.NO_CONNECTION, icepap_system, 0])

        QtWidgets.QApplication.instance().restoreOverrideCursor()

        return conflictsList

    @loggingInfo
    def checkFirmwareVersions(self, icepap_system):
        config = ConfigManager()
        if config._options.skipversioncheck is True:
            self.log.warning("Firmware versions are not checked.")
            return
        try:
            icepap_name = icepap_system.name
            ipap = self._ctrl_icepap.iPaps[icepap_name]
            master_version = ipap.fver
            mismatched_drivers = []
            for driver in icepap_system.getDrivers():
                # #############################################################
                try:
                    driver_version = ipap[driver.addr].fver
                except Exception:
                    # If the driver has been temporary removed,
                    # the error should
                    self.log.warning('%s[%d] is removed', icepap_name,
                                     driver.addr)
                    driver_version = master_version
                if master_version != driver_version:
                    mismatched_drivers.append((driver.addr,
                                               str(driver_version)))

            if len(mismatched_drivers) > 0:
                msg = "Some drivers do not have the MASTER's " \
                      "firmware version (%s):\n" % master_version
                for driver, version in mismatched_drivers:
                    msg = msg + "driver %d: %s\n" % (driver, version)
                saved_version = ipap.ver_saved.driver[0]
                msg = msg + "Board saved version: %s\n" % saved_version
                msg = msg + "Would you like to upgrade these drivers?\n"
                self.log.info(msg)
                upgrade = MessageDialogs.showYesNoMessage(
                    self._form, "Firmware mismatch", msg)
                if upgrade:
                    progress_dialog = QtWidgets.QProgressDialog(self._form)
                    progress_dialog.setLabel(QtWidgets.QLabel(
                        "Icepap: %s\nUpgrading drivers' firmware "
                        "to %s" % (icepap_name, saved_version)))
                    progress_dialog.setCancelButton(None)
                    progress_dialog.setMaximum(100)
                    upgrading = self._ctrl_icepap.upgradeDrivers(
                        icepap_name, progress_dialog)
                    if not upgrading:
                        progress_dialog.cancel()
                        msg = "Sorry, problems found while upgrading. " \
                              "Please try it manually :-("
                        self.log.error(msg)
                        MessageDialogs.showErrorMessage(
                            None, "Firmware upgrade error", msg)
        except Exception as e:
            self.log.error('Error on checking %s firmware version: %s',
                           icepap_name, e)

    @loggingInfo
    def importMovedDriver(self, icepap_driver, from_to=False):
        """ function to import the information from a moved driver,
        to avoid losing historic cfgs"""
        id = icepap_driver.current_cfg.getParameter("ID", True)
        got_driver = self._db.existsDriver(icepap_driver, id)
        if from_to:
            if got_driver:
                src_driver = icepap_driver
                dst_driver = got_driver
            else:
                delete = MessageDialogs.showYesNoMessage(
                    self._form, "Driver error",
                    "Driver not present.\nRemove driver from DB?")
                if delete:
                    icepap_driver.icepap_system.removeDriver(
                        icepap_driver.addr)
                return None
        else:
            src_driver = got_driver
            dst_driver = icepap_driver

        move = MessageDialogs.showYesNoMessage(
            self._form, "Import moved driver",
            "Import all historic configurations from %s:axis "
            "%d to %s:axis %d" % (
                src_driver.icepapsystem_name, src_driver.addr,
                dst_driver.icepapsystem_name, dst_driver.addr))
        src_sys = src_driver.icepap_system
        if move:
            for cfg in src_driver.historic_cfgs:
                cfg.setDriver(dst_driver)
                dst_driver.historic_cfgs.add(cfg)
        src_driver.icepap_system.removeDriver(src_driver.addr)
        return src_sys

    @loggingInfo
    def getDriversToSign(self):
        """ Gets the all drivers to be signed """
        signList = []
        for icepap_system in list(self.IcepapSystemList.values()):
            """ TO-DO STORM review"""
            for driver in icepap_system.getDrivers():
                if driver.mode == Mode.CONFIG:
                    signList.append(driver)
        return signList

    @loggingInfo
    def getDriverConfiguration(self, icepap_name, addr):
        try:
            config = self._ctrl_icepap.getDriverConfiguration(icepap_name,
                                                              addr)
            return config

        except Exception:
            MessageDialogs.showErrorMessage(
                self._form, "GetDriverConfiguration Icepap error",
                "%s Connection error" % icepap_name)

    @loggingInfo
    def getDriverStatus(self, icepap_name, addr):
        """ Driver Status used in the System and crate view
            Returns [status register, power status, current ]"""
        icepap_system = self.getIcepapSystem(icepap_name)
        if icepap_system is None:
            return -1, False, -1
        try:
            driver = icepap_system.getDriver(addr)
            if driver is None:
                return -1, False, -1
            if driver.conflict == Conflict.DRIVER_NOT_PRESENT or \
                    driver.conflict == Conflict.NO_CONNECTION:
                # print "OUPS! THIS DRIVER SHOULD NOT BE SCANNED"
                return -1, False, -1
            status = self._ctrl_icepap.getDriverStatus(icepap_name, addr)
            return status
        except Exception:
            msg = "{},{} Connection timeout".format(icepap_name, addr)
            self.log.error(msg)
            MessageDialogs.showErrorMessage(self._form,
                                            'GetDriverStatus Icepap error ',
                                            msg)
            self._form.refreshTree()
            return -1, False, -1

    @loggingInfo
    def getDriverTestStatus(self, icepap_name, addr, pos_sel, enc_sel):
        """ Driver Status used in the System and crate view
            Returns [status register, power state, [position register value,
            encoder register value]"""
        try:
            return self._ctrl_icepap.getDriverTestStatus(icepap_name, addr,
                                                         pos_sel, enc_sel)
        except Exception as error:
            MessageDialogs.showErrorMessage(
                self._form, "GetDriverTestStatus Icepap error",
                "%s Connection error:%s" % (icepap_name, str(error)))
            self._form.refreshTree()
            return -1, -1, [-1, -1]
        except Exception as e:
            self.log.error("getDriverTestStatus:Unexpected error while "
                           "getting driver test status. %s", e)
            return -1, -1, [-1, -1]

    @loggingInfo
    def readIcepapParameters(self, icepap_name, addr, par_list):
        """ Gets from adriver the values of the parameters in par_list """
        try:
            return self._ctrl_icepap.readIcepapParameters(icepap_name,
                                                          addr, par_list)
        except Exception as error:
            msg = "{} Connection " \
                  "error: {}".format(icepap_name, error)
            self.log.error(msg)
            MessageDialogs.showErrorMessage(
                self._form, 'ReadIcepapParamenters Icepap error ', msg)

    @loggingInfo
    def writeIcepapParameters(self, icepap_name, addr, par_var_list):
        """ Writes to a driver the values in the par_var_list """
        try:
            self._ctrl_icepap.writeIcepapParameters(icepap_name, addr,
                                                    par_var_list)
        except Exception:
            pass

    @loggingInfo
    def getDriverMotionValues(self, icepap_name, addr):
        """ Returns speed and acceleration of a driver """
        try:
            return self._ctrl_icepap.getDriverMotionValues(icepap_name, addr)
        except Exception:
            return -1, -1

    @loggingInfo
    def setDriverMotionValues(self, icepap_name, addr, values):
        """ Sets speed and acceleration to a driver """
        try:
            return self._ctrl_icepap.setDriverMotionValues(icepap_name,
                                                           addr, values)
        except Exception:
            MessageDialogs.showWarningMessage(
                self._form, "SetDriverMotionValues Icepap error",
                "Connection error")

    @loggingInfo
    def setDriverPosition(self, icepap_name, addr, pos_sel, position):
        return self._ctrl_icepap.setDriverPosition(icepap_name, addr,
                                                   pos_sel, position)

    @loggingInfo
    def setDriverEncoder(self, icepap_name, addr, pos_sel, position):
        return self._ctrl_icepap.setDriverEncoder(icepap_name, addr,
                                                  pos_sel, position)

    @loggingInfo
    def moveDriver(self, icepap_name, addr, steps):
        try:
            self._ctrl_icepap.moveDriver(icepap_name, addr, steps)
        except Exception as error:
            msg = "{} Connection error: {}".format(icepap_name, error)
            self.log.error(msg)
            MessageDialogs.showErrorMessage(self._form,
                                            'MoveDriver Icepap error', msg)

    @loggingInfo
    def moveDriverAbsolute(self, icepap_name, addr, position):
        try:
            self._ctrl_icepap.moveDriverAbsolute(icepap_name, addr, position)
        except Exception as error:
            msg = "{} Connection error: {}".format(icepap_name, error)
            self.log.error(msg)
            MessageDialogs.showErrorMessage(self._form,
                                            'MoveDriverAbsolute Icepap', msg)

    @loggingInfo
    def stopDriver(self, icepap_name, addr):
        try:
            self._ctrl_icepap.stopDriver(icepap_name, addr)
        except Exception as error:
            msg = "{} Connection error: {}".format(icepap_name, error)
            self.log.error(msg)
            MessageDialogs.showWarningMessage(self._form,
                                              'StopDriver Icepap error', msg)

    @loggingInfo
    def blinkDriver(self, icepap_name, driver_addr, secs):
        self._ctrl_icepap.blinkDriver(icepap_name, driver_addr, secs)


    @loggingInfo
    def jogDriver(self, icepap_name, addr, speed):
        self._ctrl_icepap.jogDriver(icepap_name, addr, speed)


    @loggingInfo
    def enableDriver(self, icepap_name, driver_addr):
        self._ctrl_icepap.enableDriver(icepap_name, driver_addr)


    @loggingInfo
    def disableDriver(self, icepap_name, driver_addr):
        self._ctrl_icepap.disableDriver(icepap_name, driver_addr)


    @loggingInfo
    def saveValuesInIcepap(self, icepap_driver, new_values,
                           expertFlag=False, ignore_values=[]):
        """ Stores the new configuration in the icepap, and sets the mode of
        the driver to CONFIG """
        for param, db_value in ignore_values:
            new_values.remove((param, db_value))
        try:
            new_cfg = self._ctrl_icepap.setDriverConfiguration(
                icepap_driver.icepapsystem_name, icepap_driver.addr,
                new_values, expertFlag=expertFlag)
        except Exception:
            new_cfg = None
        if new_cfg is None:
            return False
        else:
            icepap_driver.mode = str(Mode.CONFIG)
            icepap_driver.addConfiguration(new_cfg)
            return True


    @loggingInfo
    def startConfiguringDriver(self, icepap_driver):
        return self._ctrl_icepap.startConfiguringDriver(
            icepap_driver.icepapsystem_name, icepap_driver)

    @loggingInfo
    def endConfiguringDriver(self, icepap_driver):
        self._ctrl_icepap.endConfiguringDriver(icepap_driver.icepapsystem_name,
                                               icepap_driver)

    @loggingInfo
    def discardDriverChanges(self, icepap_driver):
        icepap_driver.setStartupCfg()
        self._ctrl_icepap.setDriverConfiguration(
            icepap_driver.icepapsystem_name, icepap_driver.addr,
            icepap_driver.current_cfg.toList())

        pass

    @loggingInfo
    def undoDriverConfiguration(self, icepap_driver):
        undo_cfg = icepap_driver.getUndoList()
        new_cfg = self._ctrl_icepap.setDriverConfiguration(
            icepap_driver.icepapsystem_name, icepap_driver.addr,
            undo_cfg.toList())
        if new_cfg is None:
            MessageDialogs.showWarningMessage(
                self._form, "undoDriverConfig error", "Connection error")
        else:
            icepap_driver.undo(new_cfg)
            return True

    @loggingInfo
    def getIcepapList(self):
        return self.IcepapSystemList

    @loggingInfo
    def getDriverTemplateList(self):
        return self._zodb.getAllDriverTemplate()

    @loggingInfo
    def saveDriverTemplate(self, name, desc, cfg):
        pass
        # TODO investigate IcepapDriverTemple
        # self._zodb.addDriverTemplate(IcepapDriverTemplate(name, desc, cfg))

    @loggingInfo
    def deleteDriverTemplate(self, name):
        self._zodb.deleteDriverTemplate(name)

    @loggingInfo
    def configDriverToDefaults(self, icepap_driver):
        icepap_name = icepap_driver.icepapsystem_name
        addr = icepap_driver.addr
        self._ctrl_icepap.configDriverToDefaults(icepap_name, addr)

    @loggingInfo
    def updateDriverConfig(self, icepap_driver):
        icepap_name = icepap_driver.icepapsystem_name
        addr = icepap_driver.addr
        driver_cfg = self.getDriverConfiguration(icepap_name, addr)
        icepap_driver.addConfiguration(driver_cfg)



#!/usr/bin/env python

from setuptools import setup, find_packages


long_description = """
IcepapCMS is a graphical configuration and test tool for the Icepap motor 
controller, base on icepap library. 

It allows to configure the electrical parameters, I/O interfaces, encoders, 
movement parameters, homing procedures. 

All changes are saved on a data base which allows to do easy rollbacks, 
identified changes and conflicts between the hardware configuration and the 
data base configuration. 

The program includes one graphical console to communicate with raw commands.  
"""

# The version is updated automatically with bumpversion
# Do not update manually
__version = '3.1.3'


# Setup
setup(
    name="icepapcms",
    version=__version,
    packages=find_packages(),
    description="Icepap Configuration Management System and Test Tool",
    long_description=long_description,
    author="Guifre Cuni et al.",
    author_email="ctbeamlines@cells.es",
    license="GPLv3",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.5',
        'Topic :: Communications',
        'Topic :: Software Development :: Libraries',
    ],
    entry_points={
        'gui_scripts': [
            'icepapcms = icepapcms.__main__:main',
            'ipapconsole = icepapcms.gui.ipapconsole:main'],
    },
    install_requires=[
        "storm>=0.23",
        "IPy>=0.62",
        "PyQt5",
        "icepap>=3.6.1",
        'configobj',
        'ldap3',
        'PyYAML',

    ],
    package_data={
        '': ['*.ui', '*.png', '*.svg', '*.gif', '*.xml', '*.txt', '*.sql',
             '*.pdf']
    },
    include_package_data=True,
    platforms=["Linux, Windows XP/Vista/7/8"],
    url="https://github.com/ALBA-Synchrotron/IcepapCMS",
    python_requires='>=3.5',
)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# This file is part of icepapcms https://github.com/ALBA-Synchrotron/icepapcms
#
# Copyright 2008-2018 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# -----------------------------------------------------------------------------


from storm.locals import Storm, Unicode, ReferenceSet, Int
import logging
from .conflict import Conflict
from .stormmanager import StormManager
from .icepapdriver import IcepapDriver
from ..helpers import loggingInfo

__all__ = ['IcepapSystem', 'Location']


class Location(Storm):
    __storm_table__ = "location"
    name = Unicode(primary=True)
    systems = ReferenceSet(name, "IcepapSystem.location_name")
    log = logging.getLogger('{}.Location'.format(__name__))

    @loggingInfo
    def __init__(self, name):
        self.name = name
        self.initialize()

    @loggingInfo
    def initialize(self):
        self._inmemory_systems = {}

    @loggingInfo
    def __storm_loaded__(self):
        self.initialize()

    @loggingInfo
    def loadSystemsfromDB(self):
        for system in self.systems:
            self._inmemory_systems[system.name] = system

    @loggingInfo
    def addSystem(self, system):
        self.systems.add(system)
        self._inmemory_systems[system.name] = system

    @loggingInfo
    def deleteSystem(self, name):
        system = self.systems.find(IcepapSystem.name == name).one()
        for driver in system.getDrivers():
            system.removeDriver(driver.addr)
        StormManager().deleteIcepapSystem(system)
        try:
            del self._inmemory_systems[name]
            self._inmemory_drivers = {}
        except Exception:
            pass


class IcepapSystem(Storm):
    __storm_table__ = "icepapsystem"
    name = Unicode(primary=True)
    host = Unicode()
    port = Int()
    description = Unicode()
    version = Unicode()
    location_name = Unicode()
    """ references """
    drivers = ReferenceSet(name, "IcepapDriver.icepapsystem_name")
    location = ReferenceSet(location_name, "Location.name")
    log = logging.getLogger('{}.IcepapSystem'.format(__name__))

    @loggingInfo
    def __init__(self, name, host, port, location_name, description=None):
        self.name = str(name)
        if description is None:
            description = str("")
        self.description = str(description)
        self.location_name = location_name
        self.host = str(host)
        self.port = int(port)
        self.initialize()

    @loggingInfo
    def __storm_loaded__(self):
        self.initialize()

    @loggingInfo
    def loadDriversfromDB(self):
        for driver in self.drivers:
            self._inmemory_drivers[driver.addr] = driver

    @loggingInfo
    def initialize(self):
        self._inmemory_drivers = {}
        self.conflict = Conflict.NO_CONFLICT
        self.child_conflicts = 0

    @loggingInfo
    def getDriver(self, addr, in_memory=True):
        if in_memory:
            if addr in self._inmemory_drivers:
                return self._inmemory_drivers[addr]
            else:
                return None
        else:
            return self.drivers.find(IcepapDriver.addr == addr).one()

    @loggingInfo
    def getDrivers(self, in_memory=True):
        if in_memory:
            if len(self._inmemory_drivers) == 0:
                self.loadDriversfromDB()
            return list(self._inmemory_drivers.values())
        else:
            return self.drivers

    @loggingInfo
    def addDriver(self, driver):
        self.drivers.add(driver)
        self._inmemory_drivers[driver.addr] = driver

    @loggingInfo
    def addDriverList(self, driver_list):
        for driver in list(driver_list.values()):
            self.addDriver(driver)

    @loggingInfo
    def removeDriver(self, addr):
        driver = self.drivers.find(IcepapDriver.addr == addr).one()
        StormManager().deleteDriver(driver)
        del self._inmemory_drivers[addr]

    @loggingInfo
    def setConflict(self, conflict):
        self.conflict = conflict

    @loggingInfo
    def signSystem(self):
        for driver in self.getDrivers():
            driver.signDriver()

    @loggingInfo
    def signCrate(self, cratenr):
        min_addr = cratenr * 10
        max_addr = (cratenr + 1) * 10
        for driver in self.drivers.find(IcepapDriver.addr > min_addr,
                                        IcepapDriver.addr < max_addr):
            driver.signDriver()

    @loggingInfo
    def compareDriverList(self, driver_list):
        self.child_conflicts = 0
        conflictsList = []
        self.conflict = Conflict.NO_CONFLICT
        ''' comparing drivers '''
        for driver in self.drivers:
            addr = driver.addr
            if addr not in driver_list:
                conflictsList.append([Conflict.DRIVER_NOT_PRESENT, self, addr])
                self.child_conflicts += 1
            else:
                driver_cmp = driver_list[addr]
                if driver != driver_cmp:

                    # HOOK TO CHECK AUTO-SOLVE CONFLICTS
                    dsp_cfg = driver_cmp.current_cfg
                    db_cfg = driver.current_cfg
                    conflict = self.checkAutoSolvedConflict(dsp_cfg, db_cfg)
                    conflictsList.append([conflict, self, addr])
                    self.child_conflicts += 1

        # checking for new drivers
        for addr, driver in list(driver_list.items()):
            if self.drivers.find(IcepapDriver.addr == addr).count() == 0:
                self.addDriver(driver)
                # determine if it is a new driver or if it has been moved
                # ALWAYS TREAT AS NEW DRIVER, 'MOVED' HAS NO SENSE
                conflictsList.append([Conflict.NEW_DRIVER, self, addr])

        return conflictsList

    @loggingInfo
    def checkAutoSolvedConflict(self, dsp_cfg, db_cfg):
        # 20130710 ESRF ASKED FOR A HOOK TO 'SKIP' SOME CONFLICTS
        # ISSUE 053 in WIKI MINUTES
        # http://wikiserv.esrf.fr/esl/index.php/IcePAP_minute_130708
        #
        # TWO NEW CONFLICT TYPES ADDED: DRIVER_AUTOSOLVE,
        # DRIVER_AUTOSOLVE_EXPERT
        # Since there is the possibility to keep current behaviour,
        # the method can return also DRIVER_CHANGED

        # NOTE: u'VER' and u'IPAPNAME' are also available to resolve
        # conflicts...

        # NOTE: configs have the .getParameter(par) method
        #       BUT dsp values are not stored in the database, so in_memory
        #       has to be set to True
        # ## par = u'VER'
        # ## print 'dsp', par, dsp_cfg.getParameter(par, in_memory=True)
        # ## print 'db', par, db_cfg.getParameter(par)

        # NOTE: it is also possible to operate with lists:
        # ##dsp_values = dsp_cfg.toList()
        # ##db_values = db_cfg.toList()
        # ##for p,v in dsp_values:
        # ##    if p == par:
        # ##        print 'dsp', p, v
        # ##for p,v in db_values:
        # ##    if p == par:
        # ##        print 'db', p, v

        try:
            dsp_cfg_ver = float(dsp_cfg.getParameter(str("VER"),
                                                     in_memory=True))
        except Exception:
            self.log.error("%s: missing VERsion parameter in DSP config",
                           self.name)
            return Conflict.DRIVER_CHANGED

        try:
            db_cfg_ver = float(db_cfg.getParameter(str("VER")))
        except Exception:
            self.log.error("%s: missing VERsion parameter in database config",
                           self.name)
            return Conflict.DRIVER_CHANGED

        if(dsp_cfg_ver == 2.0) and (db_cfg_ver < 2.0) and (db_cfg_ver >= 1.22):
            dsp_values = dsp_cfg.toList()
            db_values = db_cfg.toList()
            diff_values = set(dsp_values).difference(db_values)
            for p, v in diff_values:
                if p == 'VER':
                    continue

                if p not in ['EXTDISABLE', 'PCLMODE', 'EXTBUSY',
                             'POSUPDATE', 'LNKNAME', 'EXTPOWER', 'OUTPSRC']:
                    return Conflict.DRIVER_CHANGED
            return Conflict.DRIVER_AUTOSOLVE

        if(dsp_cfg_ver > 3.14) and (db_cfg_ver <= 3.14) and \
                (db_cfg_ver >= 2.0):
            dsp_values = dsp_cfg.toList()
            db_values = db_cfg.toList()
            diff_values = set(dsp_values).difference(db_values)
            for p, v in diff_values:
                # ignore new parameters or parameters that normally change
                if (p == 'VER') or (p == 'HOLDTIME') or (p == 'EXTHOLD'):
                    continue

                # parameters which value is not backward compatible
                if p not in ['INFOASRC', 'INFOBSRC', 'INFOCSRC']:
                    return Conflict.DRIVER_CHANGED

            return Conflict.DRIVER_AUTOSOLVE_EXPERT

        return Conflict.DRIVER_CHANGED



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# This file is part of icepapcms https://github.com/ALBA-Synchrotron/icepapcms
#
# Copyright 2008-2018 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# -----------------------------------------------------------------------------


from storm.locals import Storm, Int, Unicode, Reference, ReferenceSet
import time
from datetime import datetime
import socket
from icepap import Mode
import logging
from .conflict import Conflict
from .configmanager import ConfigManager
from .icepapsmanager import IcepapsManager
from .stormmanager import StormManager
from ..helpers import loggingInfo


__all__ = ['IcepapDriver']


class IcepapDriver(Storm):
    __storm_table__ = "icepapdriver"
    __storm_primary__ = ("icepapsystem_name", "addr")
    icepapsystem_name = Unicode()
    addr = Int()
    name = Unicode()
    mode = Unicode()
    """ references """
    icepap_system = Reference(icepapsystem_name, "IcepapSystem.name")
    historic_cfgs = ReferenceSet((icepapsystem_name, addr),
                                 ("IcepapDriverCfg.icepapsystem_name",
                                  "IcepapDriverCfg.driver_addr"))

    log = logging.getLogger('{}.IcepapDriver'.format(__name__))

    @loggingInfo
    def __init__(self, icepap_name, addr):
        self.icepapsystem_name = str(icepap_name)
        self.addr = addr
        self.current_cfg = None
        self.initialize()

    @loggingInfo
    def __storm_loaded__(self):
        self.current_cfg = self.historic_cfgs.order_by("date").last()
        self.initialize()

    @loggingInfo
    def initialize(self):
        self.drivernr = self.addr % 10
        self.cratenr = self.addr // 10
        self._undo_list = []
        self.startup_cfg = self.current_cfg
        self.conflict = Conflict.NO_CONFLICT

    @loggingInfo
    def addConfiguration(self, cfg, current=True):
        if current:
            if self.current_cfg is not None:
                self._undo_list.append(self.current_cfg)
            else:
                self.startup_cfg = cfg
            self.current_cfg = cfg
            cfg.setDriver(self)
        self.historic_cfgs.add(cfg)

    @loggingInfo
    def setConflict(self, conflict):
        self.conflict = conflict

    @loggingInfo
    def getName(self):
        return self.name

    @loggingInfo
    def setName(self, name):
        self.name = str(name)

    @loggingInfo
    def setMode(self, mode):
        self.mode = str(mode)

    @loggingInfo
    def signDriver(self):
        # AS ESRF SAYS, WHEN SIGNING THE DRIVER CONFIG, THE COMMIT SHOULD
        # BE DONE IN THE DATABASE FIRST, AND IF NO ERRORS, THEN COMMUNICATE
        # THE DRIVER THAT THE VALUES SHOULD BE SIGNED.
        try:
            user = ConfigManager().username
            host = socket.gethostname()
            signature = user + "@" + host + "_" + \
                datetime.now().strftime('%Y/%m/%d_%H:%M:%S')
            IcepapsManager().signDriverConfiguration(
                self.icepapsystem_name, self.addr, signature)
            self.mode = str(Mode.OPER)
            db = StormManager()
            db.commitTransaction()
            self.current_cfg.name = str(time.ctime())
            self.current_cfg.setSignature(signature)
            self.startup_cfg = self.current_cfg
            self.conflict = Conflict.NO_CONFLICT
        except Exception as e:
            self.log.error("some exception while trying to sign the driver %s",
                           e)

    @loggingInfo
    def setStartupCfg(self):
        self.current_cfg = self.startup_cfg
        self.conflict = Conflict.NO_CONFLICT

    @loggingInfo
    def undo(self, config):
        self.addConfiguration(config)
        # THE CURRENT CONFIGURATION SHOULD NOT BE IN THE UNDO LIST
        return self._undo_list.pop()

    @loggingInfo
    def getUndoList(self):
        return self._undo_list.pop()

    @loggingInfo
    def hasUndoList(self):
        return len(self._undo_list) > 0

    @loggingInfo
    def saveHistoricCfg(self, now, name, desc):
        self.current_cfg.name = str(name)
        self.current_cfg.description = str(desc)

    @loggingInfo
    def deleteHistoricCfg(self, cfg):
        self.historic_cfgs.remove(cfg)

    @loggingInfo
    def __ne__(self, other):
        return not self.__eq__(other)

    @loggingInfo
    def __eq__(self, other):
        if self.current_cfg == other.current_cfg:
            self.setConflict(Conflict.NO_CONFLICT)
            return True

        self.setConflict(Conflict.DRIVER_CHANGED)

        return False

    # TO SORT THE ICEPAP DRIVERS IN THE TREE
    @loggingInfo
    def __lt__(self, other):
        if isinstance(other, IcepapDriver):
            return self.addr < other.addr



from .mainmanager import *
from .icepapsystem import *
from .icepapdriver import *
from .conflict import Conflict
from .configmanager import ConfigManager
from .icepapsmanager import *
from .stormmanager import *



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# This file is part of icepapcms https://github.com/ALBA-Synchrotron/icepapcms
#
# Copyright 2008-2018 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# -----------------------------------------------------------------------------


from PyQt5 import QtCore, QtGui, QtWidgets
import logging
from ..helpers import loggingInfo


class PyConsoleText(QtWidgets.QTextEdit):
    commandReceived = QtCore.pyqtSignal(str, name='commandReceived')
    log = logging.getLogger('{}.PyConsoleText'.format(__name__))

    @loggingInfo
    def __init__(self, parent=None):
        QtWidgets.QTextEdit.__init__(self, parent)

        # to exit the main interpreter by a Ctrl-D if PyCute has no parent
        if parent is None:
            self.eofKey = QtCore.Qt.Key_D
        else:
            self.eofKey = None
        self.prompt = ""
        # last line + last incomplete lines
        self.line = ""
        self.lines = []
        # the cursor position in the last line
        self.point = 0
        # flag: the interpreter needs more input to run the last lines.
        self.more = 0
        # flag: readline() is being used for e.g. raw_input() and input()
        self.reading = 0
        # history
        self.history = []
        self.pointer = 0
        self.cursor_pos = 0

        # user interface setup

        self.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)

    @loggingInfo
    def setPrompt(self, text):
        self.prompt = text

    @loggingInfo
    def flush(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        pass

    @loggingInfo
    def readline(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        self.reading = 1
        self.__clearLine()
        self.moveCursor(QtGui.QTextCursor.End)
        if len(self.line) == 0:
            return '\n'
        else:
            return self.line

    @loggingInfo
    def write(self, text):

        # The output of self.append(text) contains to many newline characters,
        # so work around QTextEdit's policy for handling newline characters.

        cursor = self.textCursor()

        cursor.movePosition(QtGui.QTextCursor.End)

        pos1 = cursor.position()
        cursor.insertText(text)

        self.cursor_pos = cursor.position()
        self.setTextCursor(cursor)
        self.ensureCursorVisible()

        # Set the format
        cursor.setPosition(pos1, QtGui.QTextCursor.KeepAnchor)

    @loggingInfo
    def __run(self):
        """
        Append the last line to the history list, let the interpreter execute
        the last line(s), and clean up accounting for the interpreter results:
        (1) the interpreter succeeds
        (2) the interpreter fails, finds no errors and wants more line(s)
        (3) the interpreter fails, finds errors and writes them to sys.stderr
        """
        self.pointer = 0
        self.history.append(self.line)
        try:
            self.lines.append(self.line)
        except Exception as e:
            self.log.error('Error on __run(): %s', e)

        source = '\n'.join(self.lines)

        self.commandReceived.emit(source)
        self.lines = []
        self.__clearLine()
        self.write(self.prompt)

    @loggingInfo
    def __clearLine(self):
        """
        Clear input line buffer
        """
        self.line = ""
        self.point = 0

    @loggingInfo
    def __insertText(self, text):
        """
        Insert text at the current cursor position.
        """

        self.line = self.line[:self.point] + text + self.line[self.point:]
        self.point += len(text)

        cursor = self.textCursor()
        cursor.insertText(text)
        self.update()

    @loggingInfo
    def keyPressEvent(self, e):
        """
        Handle user input a key at a time.
        """
        text = e.text()
        key = e.key()

        if key == QtCore.Qt.Key_Backspace:
            if self.point:
                cursor = self.textCursor()
                cursor.movePosition(
                    QtGui.QTextCursor.PreviousCharacter,
                    QtGui.QTextCursor.KeepAnchor)
                cursor.removeSelectedText()

                self.point -= 1
                self.line = self.line[:self.point] + self.line[self.point+1:]

        elif key == QtCore.Qt.Key_Delete:
            cursor = self.textCursor()
            cursor.movePosition(
                QtGui.QTextCursor.NextCharacter,
                QtGui.QTextCursor.KeepAnchor)
            cursor.removeSelectedText()

            self.line = self.line[:self.point] + self.line[self.point+1:]

        elif key == QtCore.Qt.Key_Return or key == QtCore.Qt.Key_Enter:
            self.write('\n')
            if self.reading:
                self.reading = 0
            else:
                self.__run()

        elif key == QtCore.Qt.Key_Tab:
            self.__insertText(text)
        elif key == QtCore.Qt.Key_Left:
            if self.point:
                self.moveCursor(QtGui.QTextCursor.Left)
                self.point -= 1
        elif key == QtCore.Qt.Key_Right:
            if self.point < len(self.line):
                self.moveCursor(QtGui.QTextCursor.Right)
                self.point += 1

        elif key == QtCore.Qt.Key_Home:
            cursor = self.textCursor()
            cursor.setPosition(self.cursor_pos)
            self.setTextCursor(cursor)
            self.point = 0

        elif key == QtCore.Qt.Key_End:
            self.moveCursor(QtGui.QTextCursor.EndOfLine)
            self.point = len(self.line)

        elif key == QtCore.Qt.Key_Up:

            if len(self.history):
                if self.pointer == 0:
                    self.pointer = len(self.history)
                self.pointer -= 1
                self.__recall()

        elif key == QtCore.Qt.Key_Down:
            if len(self.history):
                self.pointer += 1
                if self.pointer == len(self.history):
                    self.pointer = 0
                self.__recall()

        elif len(text) > 0:
            self.__insertText(text)
            return

        else:
            e.ignore()
        self.update()

    @loggingInfo
    def __recall(self):
        """
        Display the current item from the command history.
        """
        cursor = self.textCursor()
        cursor.select(QtGui.QTextCursor.LineUnderCursor)
        cursor.removeSelectedText()
        self.__clearLine()
        self.write(self.prompt)
        self.__insertText(self.history[self.pointer])

    @loggingInfo
    def mousePressEvent(self, e):
        """
        Keep the cursor after the last prompt.
        """
        if e.button() == QtCore.Qt.LeftButton:
            self.moveCursor(QtGui.QTextCursor.End)

    @loggingInfo
    def contentsContextMenuEvent(self, ev):
        """
        Suppress the right button context menu.
        """
        pass


class SyntaxColor:
    """ Allow to color python keywords """

    keywords = {"and", "del", "from", "not", "while", "as", "elif", "global",
                "or", "with", "assert", "else", "if", "pass", "yield", "break",
                "except", "import", "print", "class", "exec", "in", "raise",
                "continue", "finally", "is", "return", "def", "for", "lambda",
                "try"}

    def __init__(self):
        pass

    def get_color(self, word):
        """ Return a color tuple (R,G,B) depending of the string word """

        stripped = word.strip()

        if stripped in self.keywords:
            return 255, 132, 0  # orange

        elif self.is_python_string(stripped):
            return 61, 120, 9  # dark green

        else:
            return 255, 255, 255

    def is_python_string(self, str):
        """ Return True if str is enclosed by a string mark """

        return False



import yaml
from icepap import IcePAPController
import logging
import time


class IcepapSnapshot:
    """
    Class to create/restore IcePAP backups based on Ethernet communication.
    """

    def __init__(self, host='', port=5000, timeout=3):
        log_name = '{0}.IcepapSnapshot'.format(__name__)
        self.log = logging.getLogger(log_name)
        self._host = host
        self._port = port
        self._ipap = IcePAPController(host, port, timeout, auto_axes=True)
        self.snapshot = {}
        self.done = 0

    def create_snapshot(self, filename):
        factor = 100 / (len(self._ipap.axes) + 2)
        _done = 0
        self.snapshot['Date'] = time.strftime('%Y/%m/%d %H:%M:%S +%z')
        self.snapshot['System'] = system = {}
        self.snapshot['Axes'] = axes = {}
        system['HOST'] = self._host
        system['PORT'] = self._port
        system['VER'] = self._ipap.ver['SYSTEM']['VER'][0]
        _done += 1
        self.done = _done * factor
        for axis in self._ipap.axes:
            axis_snapshot = AxisSnapshot(self._ipap[axis])
            axis_snapshot.create_snapshot()
            axes[axis] = axis_snapshot.snapshot
            _done += 1
            self.done = _done * factor
        self._save(filename)
        self.done = 100

    def _save(self, filename):
        with open(filename, 'w') as f:
            yaml.safe_dump(self.snapshot, f)


class AxisSnapshot:

    def __init__(self, axis):
        self.axis = axis
        self.snapshot = {}

    def create_snapshot(self):
        drv_ver = self.axis.fver
        self.snapshot['VER'] = drv_ver


        # Get Configuration
        self.snapshot['Configuration'] = dict(self.axis.get_cfg())

        # Get Operation:
        self.snapshot['Operation'] = oper = {}

        # Attributes can fail on reading, but we save it anyway with None
        # value
        attrs = ['velocity', 'name', 'acctime', 'pcloop', 'indexer', 'infoa',
                 'infob', 'infoc', 'pos', 'pos_measure', 'pos_shftenc',
                 'pos_tgtenc', 'pos_ctrlenc', 'pos_encin', 'pos_inpos',
                 'pos_absenc', 'pos_motor', 'pos_sync', 'enc', 'enc_measure',
                 'enc_shftenc', 'enc_tgtenc', 'enc_ctrlenc', 'enc_encin',
                 'enc_inpos', 'enc_absenc', 'enc_motor', 'enc_sync', 'id']

        # Attributes can fail on reading because they are on version 3.17
        # This attributes won on the backup file if the version is < 3
        v3_attrs = ['cswitch', 'velocity_min', 'velocity_max', 'ecam',
                    'outpos', 'outpaux', 'syncpos', 'syncaux']

        attrs += v3_attrs
        attrs.sort()

        for attr in attrs:
            if attr in v3_attrs and drv_ver < 3:
                continue
            for i in range(3):
                try:
                    value = self.axis.__getattribute__(attr)
                    break
                except Exception:
                    value = None
            oper[attr] = value

        # External Disable. Valid for FW < 3
        if drv_ver < 3:
            try:
                value = eval(self.axis.send_cmd('?DISDIS')[0])
            except Exception:
                value = None
            oper['DISDIS'] = value


    # def do_check(self, axes=[]):
    #     self._cfg_bkp.pop('GENERAL')
    #     sections = self._cfg_bkp.sections()
    #     sections.pop(sections.index('SYSTEM'))
    #     sections.pop(sections.index('CONTROLLER'))
    #     for axis in axes:
    #         section = 'AXIS_{0}'.format(axis)
    #         try:
    #             sections.pop(sections.index(section))
    #         except Exception:
    #             raise ValueError('There is not backup for the axis '
    #                              '{0}'.format(axis))
    #     if len(axes) > 0:
    #         for section in sections:
    #             self._cfg_bkp.pop(section)
    #     else:
    #         for section in sections:
    #             axis = int(section.split('_')[1])
    #             axes.append(axis)
    #     self.log.info('Checking IcePAP {0} axes: {1}'.format(self._host,
    #                                                          repr(axes)))
    #     self.do_backup(axes=axes, save=False, general=False)
    #     total_diff = {}
    #     if self._cfg_bkp == self._cfg_ipap:
    #         self.log.info('No differences found')
    #     else:
    #         sections = self._cfg_bkp.sections()
    #         for section in sections:
    #             diff = dict_cfg(self._cfg_bkp[section], self._cfg_ipap[
    #                 section])
    #             if len(diff) > 0:
    #                 total_diff[section] = diff
    #         self.log.info('Differences found: {0}'.format(repr(total_diff)))
    #     return total_diff
    #
    # def do_autofix(self, diff, force=False, skip_registers=[]):
    #     """
    #     Solve inconsistencies in IcePAP configuration registers.
    #
    #     :param diff: Differences dictionary.
    #     :param force: Force overwrite of `enc` and `pos` register values.
    #     :param skip_registers: List of registers to do not overwrite
    #         when loading a saved configuration.
    #     :return:
    #     """
    #     self.active_axes(force=True)
    #     time.sleep(2)
    #     sections = list(diff.keys())
    #     axes = []
    #     for section in sections:
    #         if 'AXIS_' in section:
    #             axis = int(section.split('_')[1])
    #             axes.append(axis)
    #     axes.sort()
    #     for axis in axes:
    #         section = 'AXIS_{0}'.format(axis)
    #         registers = diff[section]
    #         for register in registers:
    #             if 'ver' in register:
    #                 continue
    #             if 'cfg' in register:
    #                 continue
    #
    #             value_bkp, value_ipap = diff[section][register]
    #
    #             if UNKNOWN in value_bkp:
    #                 continue
    #
    #             # Check DISDIS configuration
    #             if register.lower() == 'disdis':
    #                 try:
    #                     if 'KeyNot' in value_bkp:
    #                         # Version backup > 3
    #                         value = diff[section]['cfg_extdisable'][0]
    #                         if value.lower() == 'none':
    #                             cmd = 'DISDIS 1'
    #                         else:
    #                             cmd = 'DISDIS 0'
    #                         self._ipap[axis].send_cmd(cmd)
    #                         self.log.info('Fixed axis {0} disdis '
    #                                       'configuration: cfg_extdisable({1})'
    #                                       ' -> {2}'.format(axis, value, cmd))
    #                     else:
    #                         # Version backup < 3:
    #                         value_bkp = eval(value_bkp)
    #                         self._ipap[axis].send_cmd('config')
    #                         value = ['Disable', 'NONE'][value_bkp]
    #                         cfg = 'EXTDISABLE {0}'.format(value)
    #                         self._ipap[axis].set_cfg(cfg)
    #                         cmd_str = 'config conf{0:03d}'.format(axis)
    #                         self._ipap[axis].send_cmd(cmd_str)
    #
    #                         self.log.info('Fixed axis {0} disdis '
    #                                       'configuration: DISDIS {1} -> '
    #                                       'cfg_extdisable {2}'
    #                                       .format(axis, value_bkp, value))
    #                 except Exception as e:
    #                     if self._ipap[axis].mode != 'oper':
    #                         self._ipap[axis].send_cmd('config')
    #                     self.log.error('Cannot fix axis {0} disdis '
    #                                    'configuration: bkp({1}) icepap({2}).'
    #                                    ' Error {3}'.format(axis, value_bkp,
    #                                                        value_ipap, e))
    #
    #             if 'KeyNot' in value_ipap or 'KeyNot' in value_bkp:
    #                 continue
    #
    #             if register in skip_registers:
    #                 self.log.warning('Skip register by user '
    #                                  '{0}'.format(register))
    #                 continue
    #
    #             if register.startswith('enc') and not force:
    #
    #                 self.log.warning('Skip axis {0} {1}: bkp({2}) '
    #                                  'icepap({3})'.format(axis, register,
    #                                                       value_bkp,
    #                                                       value_ipap))
    #                 continue
    #             if register.startswith('pos'):
    #                 self.log.warning('Skip axis {0} {1}: bkp({2}) '
    #                                  'icepap({3})'.format(axis, register,
    #                                                       value_bkp,
    #                                                       value_ipap))
    #                 continue
    #
    #             try:
    #                 value = eval(value_bkp)
    #                 if register == 'velocity':
    #                     acctime = self._ipap[axis].acctime
    #                     self._ipap[axis].velocity = value
    #                     self._ipap[axis].acctime = acctime
    #                 else:
    #                     self._ipap[axis].__setattr__(register, value)
    #
    #                 self.log.info('Fixed axis {0} {1}: bkp({2}) '
    #                               'icepap({3})'.format(axis, register,
    #                                                    value_bkp, value_ipap))
    #             except Exception as e:
    #                 self.log.error('Cannot fix axis {0} {1}: bkp({2}) '
    #                                'icepap({3}). '
    #                                'Error {4})'.format(axis, register,
    #                                                    value_bkp, value_ipap,
    #                                                    e))
    #     self.active_axes()
    #
    # def active_axes(self, axes=[], force=False):
    #     sections = self._cfg_bkp.sections()
    #     for section in sections:
    #         if section in ['GENERAL', 'SYSTEM', 'CONTROLLER']:
    #             continue
    #         axis = int(section.split('_')[1])
    #         if axes == [] or axis in axes:
    #             active = self._cfg_bkp.getboolean(section, 'ACTIVE')
    #             self._ipap[axis].send_cmd('config')
    #             if force:
    #                 cfg = 'ACTIVE YES'
    #             else:
    #                 cfg = 'ACTIVE {0}'.format(['NO', 'YES'][active])
    #             self.log.info('Axis {0}: {1}'.format(axis, cfg))
    #             self._ipap[axis].set_cfg(cfg)
    #             cmd = 'config conf{0:03d}'.format(axis)
    #             self._ipap[axis].send_cmd(cmd)
